@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-3 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -25,67 +25,116 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <syslog.h>
+#include <string.h>
+#include "atacmdnames.h"
 #include "atacmds.h"
 #include "ataprint.h"
 #include "smartctl.h"
 #include "extern.h"
 #include "utility.h"
+#include "knowndrives.h"
 
-const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.56 2003/01/17 21:47:12 ballen4705 Exp $"
-ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
+const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.97 2003/08/13 12:33:23 ballen4705 Exp $"
+ATACMDNAMES_H_CVSID ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // for passing global control variables
-extern atamainctrl *con;
+extern smartmonctrl *con;
 
-// Function for printing ASCII byte-swapped strings, skipping white
-// space. This is needed on little-endian architectures, eg Intel,
-// Alpha. If someone wants to run this on SPARC they'll need to test
-// for the Endian-ness and skip the byte swapping if it's big-endian.
-void printswap(char *in, unsigned int n){
-  unsigned int i;
-  char out[64];
+// Copies n bytes (or n-1 if n is odd) from in to out, but swaps adjacents
+// bytes.
+void swapbytes(char *out, const char *in, size_t n)
+{
+  size_t i;
 
-  // swap bytes
-  for (i=0;i<n;i+=2){
-    unsigned int j=i+1;
-    out[i]=in[j];
-    out[j]=in[i];
+  for (i = 0; i < n; i += 2) {
+    out[i]   = in[i+1];
+    out[i+1] = in[i];
+  }
+}
+
+// Copies in to out, but removes leading and trailing whitespace.
+void trim(char *out, const char *in)
+{
+  int i, first, last;
+
+  // Find the first non-space character (maybe none).
+  first = -1;
+  for (i = 0; in[i]; i++)
+    if (!isspace(in[i])) {
+      first = i;
+      break;
+    }
+
+  if (first == -1) {
+    // There are no non-space characters.
+    out[0] = '\0';
+    return;
   }
 
-  // find the end of the white space
-  for (i=0;i<n && isspace(out[i]);i++);
+  // Find the last non-space character.
+  for (i = strlen(in)-1; i >= first && isspace(in[i]); i--)
+    ;
+  last = i;
 
-  // and do the printing starting from first non-white space
-  if (n-i)
-    pout("%.*s\n",(int)(n-i),out+i);
+  strncpy(out, in+first, last-first+1);
+  out[last-first+1] = '\0';
+}
+
+// Convenience function for formatting strings from hd_driveid.
+void formatdriveidstring(char *out, const char *in, int n)
+{
+  char tmp[65];
+
+  n = n > 64 ? 64 : n;
+  swapbytes(tmp, in, n);
+  tmp[n] = '\0';
+  trim(out, tmp);
+}
+
+// Function for printing ASCII byte-swapped strings, skipping white
+// space. Please note that this is needed on both big- and
+// little-endian hardware.
+void printswap(char *output, char *in, unsigned int n){
+  formatdriveidstring(output, in, n);
+  if (*output)
+    pout("%s\n", output);
   else
     pout("[No Information Found]\n");
-
-  return;
 }
 
 void ataPrintDriveInfo (struct hd_driveid *drive){
-  int version;
+  int version, drivetype;
   const char *description;
   char unknown[64], timedatetz[64];
   unsigned short minorrev;
+  char model[64], serial[64], firm[64];
+  
 
   // print out model, serial # and firmware versions  (byte-swap ASCI strings)
   pout("Device Model:     ");
-  printswap(drive->model,40);
+  printswap(model, drive->model,40);
 
   pout("Serial Number:    ");
-  printswap(drive->serial_no,20);
+  printswap(serial, drive->serial_no,20);
 
   pout("Firmware Version: ");
-  printswap(drive->fw_rev,8);
+  printswap(firm, drive->fw_rev,8);
+
+  // See if drive is recognized
+  drivetype=lookupdrive(model, firm);
+  pout("Device is:        %s\n", drivetype<0?
+       "Not in smartctl database [for details use: -P showall]":
+       "In smartctl database [for details use: -P show]");
 
   // now get ATA version info
   version=ataVersionInfo(&description,drive, &minorrev);
 
   // unrecognized minor revision code
   if (!description){
-    sprintf(unknown,"Unrecognized. Minor revision code: 0x%02hx",minorrev);
+    if (!minorrev)
+      sprintf(unknown, "Exact ATA specification draft version not indicated");
+    else
+      sprintf(unknown,"Not recognized. Minor revision code: 0x%02hx", minorrev);
     description=unknown;
   }
   
@@ -103,6 +152,10 @@ void ataPrintDriveInfo (struct hd_driveid *drive){
   dateandtimezone(timedatetz);
   pout("Local Time is:    %s\n", timedatetz);
 
+  // Print warning message, if there is one
+  if (drivetype>=0 && knowndrives[drivetype].warningmsg)
+    pout("\n==> WARNING: %s\n\n", knowndrives[drivetype].warningmsg);
+
   if (version>=3)
     return;
   
@@ -114,62 +167,53 @@ void ataPrintDriveInfo (struct hd_driveid *drive){
 
 /*  prints verbose value Off-line data collection status byte */
 void PrintSmartOfflineStatus(struct ata_smart_values *data){
-  pout("Off-line data collection status: ");	
+  char *message=NULL;
+
+  // the final 7 bits
+  unsigned char stat=data->offline_data_collection_status & 0x7f;
   
-  switch(data->offline_data_collection_status){
+  pout("Off-line data collection status: (0x%02x)\t",
+       (int)data->offline_data_collection_status);
+    
+  switch(stat){
   case 0x00:
-  case 0x80:
-    pout("(0x%02x)\tOffline data collection activity was\n\t\t\t\t\t",
-	 (int)data->offline_data_collection_status);
-    pout("never started.\n");
-    break;
-  case 0x01:
-  case 0x81:
-    pout("(0x%02x)\tReserved.\n",
-	 (int)data->offline_data_collection_status);
+    message="never started";
     break;
   case 0x02:
-  case 0x82:
-    pout("(0x%02x)\tOffline data collection activity \n\t\t\t\t\t",
-	 (int)data->offline_data_collection_status);
-    pout("completed without error.\n");
-    break;
-  case 0x03:
-  case 0x83:
-    pout("(0x%02x)\tReserved.\n",
-	 (int)data->offline_data_collection_status);
+    message="completed without error";
     break;
   case 0x04:
-  case 0x84:
-    pout("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
-	 (int)data->offline_data_collection_status);
-    pout("suspended by an interrupting command from host.\n");
+    message="suspended by an interrupting command from host";
     break;
   case 0x05:
-  case 0x85:
-    pout("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
-	 (int)data->offline_data_collection_status);
-    pout("aborted by an interrupting command from host.\n");
+    message="aborted by an interrupting command from host";
     break;
   case 0x06:
-  case 0x86:
-    pout("(0x%02x)\tOffline data collection activity was \n\t\t\t\t\t",
-	 (int)data->offline_data_collection_status);
-    pout("aborted by the device with a fatal error.\n");
+    message="aborted by the device with a fatal error";
     break;
   default:
-    if ( ((data->offline_data_collection_status >= 0x07) &&
-	  (data->offline_data_collection_status <= 0x3f)) ||
-	 ((data->offline_data_collection_status >= 0xc0) &&
-	  (data->offline_data_collection_status <= 0xff)) )
-      pout("(0x%02x)\tVendor Specific.\n",(int)data->offline_data_collection_status);
+    if (stat >= 0x40)
+      pout("Vendor Specific.\n");
     else
-      pout("(0x%02x)\tReserved.\n",(int)data->offline_data_collection_status);
+      pout("Reserved.\n");
   }
+  
+  if (message)
+    // Off-line data collection status byte is not a reserved
+    // or vendor specific value
+    pout("Offline data collection activity was\n"
+	 "\t\t\t\t\t%s.\n", message);
+  
+  // Report on Automatic Data Collection Status.  Only IBM documents
+  // this bit.  See SFF 8035i Revision 2 for details.
+  if (data->offline_data_collection_status & 0x80)
+    pout("\t\t\t\t\tAuto Off-line Data Collection: Enabled.\n");
+  else
+    pout("\t\t\t\t\tAuto Off-line Data Collection: Disabled.\n");
+  
+  return;
 }
 
-
-
 void PrintSmartSelfExecStatus(struct ata_smart_values *data)
 {
    pout("Self-test execution status:      ");
@@ -246,38 +290,43 @@ void PrintSmartTotalTimeCompleteOffline ( struct ata_smart_values *data){
 
 
 
-void PrintSmartOfflineCollectCap(struct ata_smart_values *data)
-{
-   pout("Offline data collection\n");
-   pout("capabilities: \t\t\t (0x%02x) ",
-            (int)data->offline_data_collection_capability);
+void PrintSmartOfflineCollectCap(struct ata_smart_values *data){
+  pout("Offline data collection\n");
+  pout("capabilities: \t\t\t (0x%02x) ",
+       (int)data->offline_data_collection_capability);
+  
+  if (data->offline_data_collection_capability == 0x00){
+    pout("\tOff-line data collection not supported.\n");
+  } 
+  else {
+    pout( "%s\n", isSupportExecuteOfflineImmediate(data)?
+	  "SMART execute Offline immediate." :
+	  "No SMART execute Offline immediate.");
+    
+    pout( "\t\t\t\t\t%s\n", isSupportAutomaticTimer(data)? 
+	  "Automatic timer ON/OFF support.":
+	  "No Automatic timer ON/OFF support.");
+    
+    pout( "\t\t\t\t\t%s\n", isSupportOfflineAbort(data)? 
+	  "Abort Offline collection upon new\n\t\t\t\t\tcommand.":
+	  "Suspend Offline collection upon new\n\t\t\t\t\tcommand.");
+    
+    pout( "\t\t\t\t\t%s\n", isSupportOfflineSurfaceScan(data)? 
+	  "Offline surface scan supported.":
+	  "No Offline surface scan supported.");
+    
+    pout( "\t\t\t\t\t%s\n", isSupportSelfTest(data)? 
+	  "Self-test supported.":
+	  "No Self-test supported.");
 
-   if (data->offline_data_collection_capability == 0x00)
-   {
-      pout("\tOff-line data collection not supported.\n");
-   } 
-   else 
-   {
-      pout( "%s\n", isSupportExecuteOfflineImmediate(data)?
-              "SMART execute Offline immediate." :
-              "No SMART execute Offline immediate.");
+    pout( "\t\t\t\t\t%s\n", isSupportConveyanceSelfTest(data)? 
+	  "Conveyance Self-test supported.":
+	  "No Conveyance Self-test supported.");
 
-      pout( "\t\t\t\t\t%s\n", isSupportAutomaticTimer(data)? 
-              "Automatic timer ON/OFF support.":
-              "No Automatic timer ON/OFF support.");
-		
-      pout( "\t\t\t\t\t%s\n", isSupportOfflineAbort(data)? 
-              "Abort Offline collection upon new\n\t\t\t\t\tcommand.":
-              "Suspend Offline collection upon new\n\t\t\t\t\tcommand.");
-
-      pout( "\t\t\t\t\t%s\n", isSupportOfflineSurfaceScan(data)? 
-              "Offline surface scan supported.":
-              "No Offline surface scan supported.");
-
-      pout( "\t\t\t\t\t%s\n", isSupportSelfTest(data)? 
-              "Self-test supported.":
-              "No Self-test supported.");
-    }
+    pout( "\t\t\t\t\t%s\n", isSupportSelectiveSelfTest(data)? 
+	  "Selective Self-test supported.":
+	  "No Selective Self-test supported.");
+  }
 }
 
 
@@ -325,38 +374,32 @@ void PrintSmartErrorLogCapability ( struct ata_smart_values *data)
 
 
 
-void PrintSmartShortSelfTestPollingTime (struct ata_smart_values *data)
-{
-   if ( isSupportSelfTest(data) )
-   {
-      pout("Short self-test routine \n");
-      pout("recommended polling time: \t (%4d) minutes.\n", 
-               (int)data->short_test_completion_time);
-
-   }
-   else
-   {
-      pout("Short self-test routine \n");
-      pout("recommended polling time: \t        Not Supported.\n");
-   }
+void PrintSmartShortSelfTestPollingTime(struct ata_smart_values *data){
+  pout("Short self-test routine \n");
+  if (isSupportSelfTest(data))
+    pout("recommended polling time: \t (%4d) minutes.\n", 
+	 (int)data->short_test_completion_time);
+  else
+    pout("recommended polling time: \t        Not Supported.\n");
 }
 
-
-void PrintSmartExtendedSelfTestPollingTime ( struct ata_smart_values *data)
-{
-   if ( isSupportSelfTest(data) )
-   {
-      pout("Extended self-test routine \n");
-      pout("recommended polling time: \t (%4d) minutes.\n", 
-               (int)data->extend_test_completion_time);
-   }
-   else
-   {
-      pout("Extended self-test routine \n");
-      pout("recommended polling time: \t        Not Supported.\n");
-   }
+void PrintSmartExtendedSelfTestPollingTime(struct ata_smart_values *data){
+  pout("Extended self-test routine\n");
+  if (isSupportSelfTest(data))
+    pout("recommended polling time: \t (%4d) minutes.\n", 
+	 (int)data->extend_test_completion_time);
+  else
+    pout("recommended polling time: \t        Not Supported.\n");
 }
 
+void PrintSmartConveyanceSelfTestPollingTime(struct ata_smart_values *data){
+  pout("Conveyance self-test routine\n");
+  if (isSupportConveyanceSelfTest(data))
+    pout("recommended polling time: \t (%4d) minutes.\n", 
+	 (int)data->conveyance_test_completion_time);
+  else
+    pout("recommended polling time: \t        Not Supported.\n");
+}
 
 // onlyfailed=0 : print all attribute values
 // onlyfailed=1:  just ones that are currently failed and have prefailure bit set
@@ -364,9 +407,9 @@ void PrintSmartExtendedSelfTestPollingTime ( struct ata_smart_values *data)
 void PrintSmartAttribWithThres (struct ata_smart_values *data, 
 				struct ata_smart_thresholds *thresholds,
 				int onlyfailed){
-  int i,j;
-  long long rawvalue;
+  int i;
   int needheader=1;
+  char rawstring[64];
     
   // step through all vendor attributes
   for (i=0; i<NUMBER_ATA_SMART_ATTRIBUTES; i++){
@@ -376,7 +419,7 @@ void PrintSmartAttribWithThres (struct ata_smart_values *data,
     
     // consider only valid attributes
     if (disk->id && thre->id){
-      char *type;
+      char *type, *update;
       int failednow,failedever;
       char attributename[64];
 
@@ -384,7 +427,7 @@ void PrintSmartAttribWithThres (struct ata_smart_values *data,
       failedever= (disk->worst   <= thre->threshold);
       
       // These break out of the loop if we are only printing certain entries...
-      if (onlyfailed==1 && (!disk->status.flag.prefailure || !failednow))
+      if (onlyfailed==1 && (!ATTRIBUTE_FLAGS_PREFAILURE(disk->flags) || !failednow))
 	continue;
       
       if (onlyfailed==2 && !failedever)
@@ -396,70 +439,33 @@ void PrintSmartAttribWithThres (struct ata_smart_values *data,
 	  pout("SMART Attributes Data Structure revision number: %d\n",(int)data->revnumber);
 	  pout("Vendor Specific SMART Attributes with Thresholds:\n");
 	}
-	pout("ID# ATTRIBUTE_NAME          FLAG     VALUE WORST THRESH TYPE     WHEN_FAILED RAW_VALUE\n");
+	pout("ID# ATTRIBUTE_NAME          FLAG     VALUE WORST THRESH TYPE      UPDATED  WHEN_FAILED RAW_VALUE\n");
 	needheader=0;
       }
       
-      // is this currently failed, or has it ever failed?
+      // is this Attribute currently failed, or has it ever failed?
       if (failednow)
 	status="FAILING_NOW";
       else if (failedever)
 	status="In_the_past";
       else
 	status="    -";
-      
+
       // Print name of attribute
       ataPrintSmartAttribName(attributename,disk->id, con->attributedefs);
       pout("%-28s",attributename);
 
       // printing line for each valid attribute
-      type=disk->status.flag.prefailure?"Pre-fail":"Old_age";
-      pout("0x%04x   %.3d   %.3d   %.3d    %-9s%-12s", 
-	     (int)disk->status.all, (int)disk->current, (int)disk->worst,
-	     (int)thre->threshold, type, status);
-      
-      // convert the six individual bytes to a long long (8 byte) integer
-      rawvalue = 0;
-      for (j=0; j<6; j++) {
-	// This looks a bit roundabout, but is necessary.  Don't
-	// succumb to the temptation to use raw[j]<<(8*j) since under
-	// the normal rules this will be promoted to the native type.
-	// On a 32 bit machine this might then overflow.
-	long long temp;
-	temp = disk->raw[j];
-	temp <<= 8*j;
-	rawvalue |= temp;
-      }
+      type=ATTRIBUTE_FLAGS_PREFAILURE(disk->flags)?"Pre-fail":"Old_age";
+      update=ATTRIBUTE_FLAGS_ONLINE(disk->flags)?"Always":"Offline";
 
-      // This switch statement is where we handle Raw attributes
-      // that are stored in an unusual vendor-specific format,
-      switch (disk->id){
-	// Power on time
-      case 9:
-	if (con->attributedefs[9]==1){
-	  // minutes
-	  long long tmp1=rawvalue/60;
-	  long long tmp2=rawvalue%60;
-	  pout("%lluh+%02llum\n", tmp1, tmp2);
-	}
-	else
-	  // hours
-	  pout("%llu\n", rawvalue);  //stored in hours
-	break;
-	
-	// Temperature
-      case 194:
-	pout("%d", (int)disk->raw[0]);
-	if (rawvalue==disk->raw[0])
-	  pout("\n");
-	else
-	  // The other bytes are in use. Try IBM's model
-	  pout(" (Lifetime Min/Max %d/%d)\n",(int)disk->raw[2],
-		 (int)disk->raw[4]);
-	break;
-      default:
-	pout("%llu\n", rawvalue);
-      }
+      pout("0x%04x   %.3d   %.3d   %.3d    %-10s%-9s%-12s", 
+	     (int)disk->flags, (int)disk->current, (int)disk->worst,
+	     (int)thre->threshold, type, update, status);
+
+      // print raw value of attribute
+      ataPrintSmartAttribRawValue(rawstring, disk, con->attributedefs);
+      pout("%s\n", rawstring);
       
       // print a warning if there is inconsistency here!
       if (disk->id != thre->id){
@@ -474,8 +480,7 @@ void PrintSmartAttribWithThres (struct ata_smart_values *data,
   if (!needheader) pout("\n");
 }
 
-
-void ataPrintGeneralSmartValues(struct ata_smart_values *data){
+void ataPrintGeneralSmartValues(struct ata_smart_values *data, struct hd_driveid *drive){
   pout("General SMART Values:\n");
   
   PrintSmartOfflineStatus(data); 
@@ -489,10 +494,18 @@ void ataPrintGeneralSmartValues(struct ata_smart_values *data){
   PrintSmartCapability(data);
   
   PrintSmartErrorLogCapability(data);
+
+  pout( "\t\t\t\t\t%s\n", isGeneralPurposeLoggingCapable(drive)?
+	"General Purpose Logging supported.":
+	"No General Purpose Logging support.");
+
   if (isSupportSelfTest(data)){
     PrintSmartShortSelfTestPollingTime (data);
     PrintSmartExtendedSelfTestPollingTime (data);
   }
+  if (isSupportConveyanceSelfTest(data))
+    PrintSmartConveyanceSelfTestPollingTime (data);
+  
   pout("\n");
 }
 
@@ -505,8 +518,63 @@ int nonempty(unsigned char *testarea,int n){
   return 0;
 }
 
+int ataPrintLogDirectory(struct ata_smart_log_directory *data){
+  int i;
+  char *name;
+
+  pout("SMART Log Directory Logging Version %d%s\n",
+       data->logversion, data->logversion==1?" [multi-sector log support]":"");
+  for (i=0; i<=255; i++){
+    int numsect;
+    
+    // Directory log length
+    numsect = i? data->entry[i-1].numsectors : 1;
+    
+    // If the log is not empty, what is it's name
+    if (numsect){
+      switch (i) {
+      case 0:
+	name="Log Directory"; break;
+      case 1:
+	name="Summary SMART error log"; break;
+      case 2:
+	name="Comprehensive SMART error log"; break;
+      case 3:
+	name="Extended Comprehensive SMART error log"; break;
+      case 6:
+	name="SMART self-test log"; break;
+      case 7:
+	name="Extended self-test log"; break;
+      case 9:
+	name="Selective self-test log"; break;
+      case 0x20:
+	name="Streaming performance log"; break;
+      case 0x21:
+	name="Write stream error log"; break;
+      case 0x22:
+	name="Read stream error log"; break;
+      case 0x23:
+	name="Delayed sector log"; break;
+      default:
+	if (0xa0<=i && i<=0xbf) 
+	  name="Device vendor specific log";
+	else if (0x80<=i && i<=0x9f)
+	  name="Host vendor specific log";
+	else
+	  name="Reserved log";
+	break;
+      }
+
+      // print name and length of log
+      pout("Log at address 0x%02x has %03d sectors [%s]\n",
+	   i, numsect, name);
+    }
+  }
+  return 0;
+}
+
 // returns number of errors
-int ataPrintSmartErrorlog (struct ata_smart_errorlog *data){
+int ataPrintSmartErrorlog(struct ata_smart_errorlog *data){
   int i,j,k;
   
   pout("SMART Error Log Version: %d\n", (int)data->revnumber);
@@ -519,7 +587,7 @@ int ataPrintSmartErrorlog (struct ata_smart_errorlog *data){
   QUIETON(con);
   // If log pointer out of range, return
   if (data->error_log_pointer>5){
-    pout("Invalid Error Log index = %02x (T13/1321D rev 1c"
+    pout("Invalid Error Log index = 0x%02x (T13/1321D rev 1c "
 	 "Section 8.41.6.8.2.2 gives valid range from 1 to 5)\n\n",
 	 (int)data->error_log_pointer);
     return 0;
@@ -538,69 +606,80 @@ int ataPrintSmartErrorlog (struct ata_smart_errorlog *data){
     pout( "ATA Error Count: %d (device log contains only the most recent five errors)\n",
 	   (int)data->ata_error_count);
   QUIETOFF(con);
-  pout("\tDCR = Device Control Register\n");
-  pout("\tFR  = Features Register\n");
-  pout("\tSC  = Sector Count Register\n");
-  pout("\tSN  = Sector Number Register\n");
-  pout("\tCL  = Cylinder Low Register\n");
-  pout("\tCH  = Cylinder High Register\n");
-  pout("\tD/H = Device/Head Register\n");
-  pout("\tCR  = Content written to Command Register\n");
-  pout("\tER  = Error register\n");
-  pout("\tSTA = Status register\n");
-  pout("Timestamp is seconds since the previous disk power-on.\n");
-  pout("Note: timestamp \"wraps\" after 2^32 msec = 49.710 days.\n\n");
+  pout("\tCR = Command Register [HEX]\n"
+       "\tFR = Features Register [HEX]\n"
+       "\tSC = Sector Count Register [HEX]\n"
+       "\tSN = Sector Number Register [HEX]\n"
+       "\tCL = Cylinder Low Register [HEX]\n"
+       "\tCH = Cylinder High Register [HEX]\n"
+       "\tDH = Device/Head Register [HEX]\n"
+       "\tDC = Device Command Register [HEX]\n"
+       "\tER = Error register [HEX]\n"
+       "\tST = Status register [HEX]\n"
+       "Timestamp = decimal seconds since the previous disk power-on.\n"
+       "Note: timestamp \"wraps\" after 2^32 msec = 49.710 days.\n\n");
   
   // now step through the five error log data structures (table 39 of spec)
   for (k = 4; k >= 0; k-- ) {
-    
+
     // The error log data structure entries are a circular buffer
     i=(data->error_log_pointer+k)%5;
     
     // Spec says: unused error log structures shall be zero filled
     if (nonempty((unsigned char*)&(data->errorlog_struct[i]),sizeof(data->errorlog_struct[i]))){
+      // Table 57 of T13/1532D Volume 1 Revision 3
       char *msgstate;
-      switch (data->errorlog_struct[i].error_struct.state){
+      int bits=data->errorlog_struct[i].error_struct.state & 0x0f;
+      switch (bits){
       case 0x00: msgstate="in an unknown state";break;
       case 0x01: msgstate="sleeping"; break;
       case 0x02: msgstate="in standby mode"; break;
       case 0x03: msgstate="active or idle"; break;
       case 0x04: msgstate="doing SMART off-line or self test"; break;
-      default:   msgstate="in a vendor specific or reserved state";
+      default:   
+	if (bits<0x0b)
+	  msgstate="in a reserved state";
+	else
+	  msgstate="in a vendor specific state";
       }
+
       // See table 42 of ATA5 spec
       QUIETON(con);
       pout("Error %d occurred at disk power-on lifetime: %d hours\n",
 	     (int)(data->ata_error_count+k-4), (int)data->errorlog_struct[i].error_struct.timestamp);
       QUIETOFF(con);
-      pout("When the command that caused the error occurred, the device was %s.\n",msgstate);
-      pout("After command completion occurred, registers were:\n");
-      pout("ER:%02x SC:%02x SN:%02x CL:%02x CH:%02x D/H:%02x ST:%02x\n",
+      pout("  When the command that caused the error occurred, the device was %s.\n\n",msgstate);
+      pout("  After command completion occurred, registers were:\n"
+	   "  ER ST SC SN CL CH DH\n"
+	   "  -- -- -- -- -- -- --\n"
+	   "  %02x %02x %02x %02x %02x %02x %02x\n\n",
 	   (int)data->errorlog_struct[i].error_struct.error_register,
+	   (int)data->errorlog_struct[i].error_struct.status,
 	   (int)data->errorlog_struct[i].error_struct.sector_count,
 	   (int)data->errorlog_struct[i].error_struct.sector_number,
 	   (int)data->errorlog_struct[i].error_struct.cylinder_low,
 	   (int)data->errorlog_struct[i].error_struct.cylinder_high,
-	   (int)data->errorlog_struct[i].error_struct.drive_head,
-	   (int)data->errorlog_struct[i].error_struct.status);
-      pout("Sequence of commands leading to the command that caused the error were:\n");
-      pout("DCR   FR   SC   SN   CL   CH   D/H   CR   Timestamp\n");
+	   (int)data->errorlog_struct[i].error_struct.drive_head);
+      pout("  Commands leading to the command that caused the error were:\n"
+	   "  CR FR SC SN CL CH DH DC   Timestamp  Command/Feature_Name\n"
+	   "  -- -- -- -- -- -- -- --   ---------  --------------------\n");
       for ( j = 4; j >= 0; j--){
 	struct ata_smart_errorlog_command_struct *thiscommand=&(data->errorlog_struct[i].commands[j]);
 	
 	// Spec says: unused data command structures shall be zero filled
 	if (nonempty((unsigned char*)thiscommand,sizeof(*thiscommand)))
-	  pout(" %02x   %02x   %02x   %02x   %02x   %02x    %02x   %02x     %d.%03d\n", 
-	       (int)thiscommand->devicecontrolreg,
+	  pout("  %02x %02x %02x %02x %02x %02x %02x %02x %7d.%03d  %s\n",
+	       (int)thiscommand->commandreg,
 	       (int)thiscommand->featuresreg,
 	       (int)thiscommand->sector_count,
 	       (int)thiscommand->sector_number,
 	       (int)thiscommand->cylinder_low,
 	       (int)thiscommand->cylinder_high,
 	       (int)thiscommand->drive_head,
-	       (int)thiscommand->commandreg,
+	       (int)thiscommand->devicecontrolreg,
 	       (unsigned int)(thiscommand->timestamp / 1000U),
-	       (unsigned int)(thiscommand->timestamp % 1000U)); 
+	       (unsigned int)(thiscommand->timestamp % 1000U),
+	       look_up_ata_command(thiscommand->commandreg, thiscommand->featuresreg));
       }
       pout("\n");
     }
@@ -618,9 +697,9 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
   int retval=0;
 
   if (allentries)
-    pout("SMART Self-test log, version number %d\n",(int)data->revnumber);
-  if ((data->revnumber!=0x0001) && allentries)
-    pout("Warning - structure revision number does not match spec!\n");
+    pout("SMART Self-test log structure revision number %d\n",(int)data->revnumber);
+  if ((data->revnumber!=0x0001) && allentries && con->fixfirmwarebug != FIX_SAMSUNG)
+    pout("Warning: ATA Specification requires self-test log structure revision number = 1\n");
   if (data->mostrecenttest==0){
     if (allentries)
       pout("No self-tests have been logged\n\n");
@@ -643,15 +722,24 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
       case   0: msgtest="Off-line           "; break;
       case   1: msgtest="Short off-line     "; break;
       case   2: msgtest="Extended off-line  "; break;
+      case   3: msgtest="Conveyance off-line"; break;
+      case   4: msgtest="Selective off-line "; break;
       case 127: msgtest="Abort off-line test"; break;
       case 129: msgtest="Short captive      "; break;
       case 130: msgtest="Extended captive   "; break;
-      default:  msgtest="Unknown test       ";
+      case 131: msgtest="Conveyance captive "; break;
+      case 132: msgtest="Selective captive  "; break;
+      default:  
+	if ( log->selftestnumber>=192 ||
+	    (log->selftestnumber>= 64 && log->selftestnumber<=126))
+	  msgtest="Vendor off-line    ";
+	else
+	  msgtest="Reserved off-line  ";
       }
       
       // test status
       switch((log->selfteststatus)>>4){
-      case  0:msgstat="Completed                    "; break;
+      case  0:msgstat="Completed without error      "; break;
       case  1:msgstat="Aborted by host              "; break;
       case  2:msgstat="Interrupted (host reset)     "; break;
       case  3:msgstat="Fatal or unknown error       "; errorfound=1; break;
@@ -659,8 +747,9 @@ int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data,int allentries){
       case  5:msgstat="Completed: electrical failure"; errorfound=1; break;
       case  6:msgstat="Completed: servo/seek failure"; errorfound=1; break;
       case  7:msgstat="Completed: read failure      "; errorfound=1; break;
-      case 15:msgstat="Test in progress             "; break;
-      default:msgstat="Unknown test status          ";
+      case  8:msgstat="Completed: handling damage?? "; errorfound=1; break;
+      case 15:msgstat="Self-test routine in progress"; break;
+      default:msgstat="Unknown/reserved test status ";
       }
 
       retval+=errorfound;
@@ -705,7 +794,7 @@ void ataPseudoCheckSmart ( struct ata_smart_values *data,
   for (i = 0 ; i < NUMBER_ATA_SMART_ATTRIBUTES ; i++) {
     if (data->vendor_attributes[i].id &&   
 	thresholds->thres_entries[i].id &&
-	data->vendor_attributes[i].status.flag.prefailure &&
+	ATTRIBUTE_FLAGS_PREFAILURE(data->vendor_attributes[i].flags) &&
 	(data->vendor_attributes[i].current <= thresholds->thres_entries[i].threshold) &&
 	(thresholds->thres_entries[i].threshold != 0xFE)){
       pout("Attribute ID %d Failed\n",(int)data->vendor_attributes[i].id);
@@ -740,7 +829,7 @@ void failuretest(int type, int returnvalue){
     exit(returnvalue);
   }
 
-  fprintf(stderr,"Smartctl internal error in failuretest(type=%d). Please contact %s\n",type,PROJECTHOME);
+  pout("Smartctl internal error in failuretest(type=%d). Please contact %s\n",type,PROJECTHOME);
   exit(returnvalue|FAILCMD);
 }
 
@@ -776,7 +865,24 @@ int ataPrintMain (int fd){
     pout("Smartctl: Hard Drive Read Identity Failed\n\n");
     failuretest(MANDATORY_CMD, returnval|=FAILID);
   }
-  
+
+  // If requested, show which presets would be used for this drive and exit.
+  if (con->showpresets) {
+    showpresets(&drive);
+    exit(0);
+  }
+
+  // Use preset vendor attribute options unless user has requested otherwise.
+  if (!con->ignorepresets){
+    unsigned char *charptr;
+    if ((charptr=con->attributedefs))
+      applypresets(&drive, &charptr, con);
+    else {
+      pout("Fatal internal error in ataPrintMain()\n");
+      exit(1);
+    }
+  }
+
   // Print most drive identity information if requested
   if (con->driveinfo){
     pout("=== START OF INFORMATION SECTION ===\n");
@@ -911,10 +1017,30 @@ int ataPrintMain (int fd){
   
   // Check SMART status (use previously returned value)
   if (con->checksmart){
-    if (code) {
+    switch (code) {
+
+    case 0:
+      // The case where the disk health is OK
+      pout("SMART overall-health self-assessment test result: PASSED\n");
+      if (ataCheckSmart(&smartval, &smartthres,0)){
+	if (con->smartvendorattrib)
+	  pout("See vendor-specific Attribute list for marginal Attributes.\n\n");
+	else {
+	  QUIETON(con);
+	  pout("Please note the following marginal Attributes:\n");
+	  PrintSmartAttribWithThres(&smartval, &smartthres,2);
+	} 
+	returnval|=FAILAGE;
+      }
+      else
+	pout("\n");
+      break;
+      
+    case 1:
+      // The case where the disk health is NOT OK
       QUIETON(con);
       pout("SMART overall-health self-assessment test result: FAILED!\n"
-	     "Drive failure expected in less than 24 hours. SAVE ALL DATA.\n");
+	   "Drive failure expected in less than 24 hours. SAVE ALL DATA.\n");
       QUIETOFF(con);
       if (ataCheckSmart(&smartval, &smartthres,1)){
 	returnval|=FAILATTR;
@@ -930,35 +1056,79 @@ int ataPrintMain (int fd){
 	pout("No failed Attributes found.\n\n");   
       returnval|=FAILSTATUS;
       QUIETOFF(con);
-    }
-    else {
-      pout("SMART overall-health self-assessment test result: PASSED\n");
-      if (ataCheckSmart(&smartval, &smartthres,0)){
+      break;
+
+    case -1:
+    default:
+      // The case where something went wrong with HDIO_DRIVE_TASK ioctl()
+      if (ataCheckSmart(&smartval, &smartthres,1)){
+	QUIETON(con);
+	pout("SMART overall-health self-assessment test result: FAILED!\n"
+	     "Drive failure expected in less than 24 hours. SAVE ALL DATA.\n");
+	QUIETOFF(con);
+	returnval|=FAILATTR;
+	returnval|=FAILSTATUS;
 	if (con->smartvendorattrib)
-	  pout("See vendor-specific Attribute list for marginal Attributes.\n\n");
+	  pout("See vendor-specific Attribute list for failed Attributes.\n\n");
 	else {
 	  QUIETON(con);
-	  pout("Please note the following marginal Attributes:\n");
-	  PrintSmartAttribWithThres(&smartval, &smartthres,2);
-	} 
-	returnval|=FAILAGE;
+	  pout("Failed Attributes:\n");
+	  PrintSmartAttribWithThres(&smartval, &smartthres,1);
+	}
       }
-      else
-	pout("\n");
-    }
+      else {
+	pout("SMART overall-health self-assessment test result: PASSED\n");
+	if (ataCheckSmart(&smartval, &smartthres,0)){
+	  if (con->smartvendorattrib)
+	    pout("See vendor-specific Attribute list for marginal Attributes.\n\n");
+	  else {
+	    QUIETON(con);
+	    pout("Please note the following marginal Attributes:\n");
+	    PrintSmartAttribWithThres(&smartval, &smartthres,2);
+	  } 
+	  returnval|=FAILAGE;
+	}
+	else
+	  pout("\n");
+      } 
+      QUIETOFF(con);
+      break;
+    } // end of switch statement
+    
     QUIETOFF(con);
-  }
+  } // end of checking SMART Status
   
   // Print general SMART values
   if (con->generalsmartvalues)
-    ataPrintGeneralSmartValues(&smartval); 
-  
+    ataPrintGeneralSmartValues(&smartval, &drive); 
+
   // Print vendor-specific attributes
   if (con->smartvendorattrib){
     QUIETON(con);
     PrintSmartAttribWithThres(&smartval, &smartthres,con->quietmode?2:0);
     QUIETOFF(con);
   }
+
+  // Print SMART log Directory
+  if (con->smartlogdirectory){
+    struct ata_smart_log_directory smartlogdirectory;
+    if (!isGeneralPurposeLoggingCapable(&drive)){
+      pout("Warning: device does not support General Purpose Logging\n");
+      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+    }
+    else {
+      QUIETON(con);
+      pout("Log Directory Supported\n");
+      if (ataReadLogDirectory(fd, &smartlogdirectory)){
+	QUIETOFF(con);
+	pout("Read Log Directory failed.\n\n");
+	failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+      }
+      else
+	ataPrintLogDirectory( &smartlogdirectory);
+    }
+    QUIETOFF(con);
+  }
   
   // Print SMART error log
   if (con->smarterrorlog){
@@ -966,20 +1136,29 @@ int ataPrintMain (int fd){
       pout("Warning: device does not support Error Logging\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
     }
-    if (ataReadErrorLog(fd, &smarterror)){
-      pout("Smartctl: SMART Errorlog Read Failed\n");
-      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
-    }
     else {
-      // quiet mode is turned on inside ataPrintSmartErrorLog()
-      if (ataPrintSmartErrorlog(&smarterror))
-	returnval|=FAILERR;
-      QUIETOFF(con);
+      if (ataReadErrorLog(fd, &smarterror)){
+	pout("Smartctl: SMART Errorlog Read Failed\n");
+	failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+      }
+      else {
+	// quiet mode is turned on inside ataPrintSmartErrorLog()
+	if (ataPrintSmartErrorlog(&smarterror))
+	  returnval|=FAILERR;
+	QUIETOFF(con);
+      }
     }
   }
   
   // Print SMART self-test log
   if (con->smartselftestlog){
+    // Note that in spite of its name, isSmartErrorLogCapable() is
+    // the CORRECT way to see if a device supports the self-test log.
+    // The ATA spec says "if this command (READ LOG) is implemented,
+    // all address values for which the contents are defined shall be
+    // implemented...".  Since both the SMART self-test logs AND the
+    // SMART error logs are defined, if one will work then so will the
+    // other.
     if (!isSmartErrorLogCapable(&smartval)){
       pout("Warning: device does not support Self Test Logging\n");
       failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
@@ -1005,14 +1184,45 @@ int ataPrintMain (int fd){
   
   pout("=== START OF OFFLINE IMMEDIATE AND SELF-TEST SECTION ===\n");
   // if doing a self-test, be sure it's supported by the hardware
-  if (con->testcase==OFFLINE_FULL_SCAN &&  !isSupportExecuteOfflineImmediate(&smartval)){
-    pout("Warning: device does not support Execute Off-Line Immediate function.\n\n");
-    failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
-  }
-  else if (!isSupportSelfTest(&smartval)){
-    pout("Warning: device does not support Self-Test functions.\n\n");
-    failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+  switch (con->testcase){
+  case OFFLINE_FULL_SCAN:
+    if (!isSupportExecuteOfflineImmediate(&smartval)){
+      pout("Warning: device does not support Execute Off-Line Immediate function.\n\n");
+      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+    }
+    break;
+  case ABORT_SELF_TEST:
+  case SHORT_SELF_TEST:
+  case EXTEND_SELF_TEST:
+  case SHORT_CAPTIVE_SELF_TEST:
+  case EXTEND_CAPTIVE_SELF_TEST:
+    if (!isSupportSelfTest(&smartval)){
+      pout("Warning: device does not support Self-Test functions.\n\n");
+      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+    }
+    break;
+  case CONVEYANCE_SELF_TEST:
+  case CONVEYANCE_CAPTIVE_SELF_TEST:
+    if (!isSupportConveyanceSelfTest(&smartval)){
+      pout("Warning: device does not support Conveyance Self-Test functions.\n\n");
+      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+    }
+    break;
+#if DEVELOP_SELECTIVE_SELF_TEST
+  case SELECTIVE_SELF_TEST:
+  case SELECTIVE_CAPTIVE_SELF_TEST:
+    if (!isSupportSelectiveSelfTest(&smartval)){
+      pout("Warning: device does not support Selective Self-Test functions.\n\n");
+      failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+    }
+    break;
+#endif
+  default:
+    pout("Internal error in smartctl: con->testcase==%d not recognized\n", (int)con->testcase);
+    pout("Please contact smartmontools developers.\n");
+    exit(returnval|=FAILCMD);
   }
+
   // Now do the test.  Note ataSmartTest prints its own error/success
   // messages
   if (ataSmartTest(fd, con->testcase))
@@ -1037,7 +1247,10 @@ int ataPrintMain (int fd){
     pout("Please wait %d %s for test to complete.\n",
 	 (int)timewait, con->testcase==OFFLINE_FULL_SCAN?"seconds":"minutes");
     
-    if (con->testcase!=SHORT_CAPTIVE_SELF_TEST && con->testcase!=EXTEND_CAPTIVE_SELF_TEST)
+    if (con->testcase!=SHORT_CAPTIVE_SELF_TEST && 
+	con->testcase!=EXTEND_CAPTIVE_SELF_TEST && 
+	con->testcase!=CONVEYANCE_CAPTIVE_SELF_TEST && 
+	con->testcase!=SELECTIVE_CAPTIVE_SELF_TEST)
       pout("Use smartctl -X to abort test.\n");	
   }    
   return returnval;
