@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-3 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -26,7 +26,7 @@
 #define _SMART_PRINT_H_
 
 #ifndef ATAPRINT_H_CVSID
-#define ATAPRINT_H_CVSID "$Id: ataprint.h,v 1.16 2003/01/17 21:47:17 ballen4705 Exp $\n"
+#define ATAPRINT_H_CVSID "$Id: ataprint.h,v 1.20 2003/04/18 12:37:14 ballen4705 Exp $\n"
 #endif
 
 #include <stdio.h>
@@ -42,25 +42,28 @@
 /* Prints ATA Drive Information and S.M.A.R.T. Capability */
 void ataPrintDriveInfo(struct hd_driveid *);
 
-void ataPrintGeneralSmartValues(struct ata_smart_values *);
+void ataPrintGeneralSmartValues(struct ata_smart_values *, struct hd_driveid *);
 
 void ataPrintSmartThresholds(struct ata_smart_thresholds *);
 
 // returns number of errors in Errorlog
 int  ataPrintSmartErrorlog(struct ata_smart_errorlog *);
 
-void PrintSmartAttributes(struct ata_smart_values *data);
+int ataPrintLogDirectory(struct ata_smart_log_directory *);
 
-void PrintSmartAttribWithThres(struct ata_smart_values *data,
-                                struct ata_smart_thresholds *thresholds,
+void PrintSmartAttributes(struct ata_smart_values *);
+
+void PrintSmartAttribWithThres(struct ata_smart_values *,
+                                struct ata_smart_thresholds *,
 				int onlyfailed);
 
 // returns number of entries that had logged errors
-int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *data, int allentries);
+int ataPrintSmartSelfTestlog(struct ata_smart_selftestlog *, int allentries);
 
 void ataPseudoCheckSmart(struct ata_smart_values *, struct ata_smart_thresholds *);
 
-
+// Convenience function for formatting strings from hd_driveid.
+void formatdriveidstring(char *out, const char *in, int n);
 
 int ataPrintMain(int fd);
 
