@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-3 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -27,17 +27,25 @@
 
 
 #ifndef EXTERN_H_CVSID
-#define EXTERN_H_CVSID "$Id: extern.h,v 1.13 2003/01/16 15:28:57 ballen4705 Exp $\n"
+#define EXTERN_H_CVSID "$Id: extern.h,v 1.26 2003/08/04 12:58:40 ballen4705 Exp $\n"
 #endif
 
+// For development and testing of Selective self-test code
+#define DEVELOP_SELECTIVE_SELF_TEST 0
+
+// Possible values for fixfirmwarebug
+#define FIX_NONE             0
+#define FIX_SAMSUNG          1
+
 // Block used for global control/communications.  If you need more
 // global variables, this should be the only place that you need to
 // add them.
-typedef struct ataprintmain_s {
+typedef struct smartmonctrl_s {
   unsigned char driveinfo;
   unsigned char checksmart;
   unsigned char smartvendorattrib;
   unsigned char generalsmartvalues;
+  unsigned char smartlogdirectory;
   unsigned char smartselftestlog;
   unsigned char smarterrorlog;
   unsigned char smartdisable;
@@ -46,8 +54,16 @@ typedef struct ataprintmain_s {
   unsigned char smartexeoffimmediate;
   unsigned char smartshortselftest;
   unsigned char smartextendselftest;
+  unsigned char smartconveyanceselftest;
+#if DEVELOP_SELECTIVE_SELF_TEST
+  unsigned char smartselectiveselftest;
+#endif
   unsigned char smartshortcapselftest;
   unsigned char smartextendcapselftest;
+  unsigned char smartconveyancecapselftest;
+#if DEVELOP_SELECTIVE_SELF_TEST
+  unsigned char smartselectivecapselftest;
+#endif
   unsigned char smartselftestabort;
   unsigned char smartautoofflineenable;
   unsigned char smartautoofflinedisable;
@@ -60,23 +76,21 @@ typedef struct ataprintmain_s {
   unsigned char conservative;
   unsigned char checksumfail;
   unsigned char checksumignore;
+  unsigned char reportataioctl;
+  unsigned char reportscsiioctl;
+  unsigned char fixfirmwarebug;
+  // If nonzero, escalade is 1 plus the disk number behind an escalade
+  // controller
+  unsigned char escalade;
+  unsigned char ignorepresets;
+  unsigned char showpresets;
   // The i'th entry in this array will modify the printed meaning of
   // the i'th SMART attribute.  The default definitions of the
   // Attributes are obtained by having the array be all zeros.  If
   // attributedefs[i] is nonzero, it means that the i'th attribute has
-  // a non-default meaning.  See the ataPrintSmartAttribName function,
-  // and list below.
+  // a non-default meaning.  See the ataPrintSmartAttribName and
+  // and parse_attribute_def functions.
   unsigned char attributedefs[256];
-} atamainctrl;
+} smartmonctrl;
 
 #endif
-
-/*
-
-attributedefs[9]:
-  1 -- time in minutes
-  2 -- temperature in Celsius
-attributedefs[220]:
-  1 -- temperature in Celsius
-
-*/
