@@ -3,9 +3,12 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-3 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
+ * Additional SCSI work:
+ * Copyright (C) 2003 Douglas Gilbert <dougg@torque.net>
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
@@ -28,9 +31,13 @@
 #define SCSI_PRINT_H_
 
 #ifndef SCSIPRINT_H_CVSID
-#define SCSIPRINT_H_CVSID "$Id: scsiprint.h,v 1.9 2003/01/16 15:28:57 ballen4705 Exp $\n"
+#define SCSIPRINT_H_CVSID "$Id: scsiprint.h,v 1.14 2003/06/17 06:06:33 dpgilbert Exp $\n"
 #endif
 
-void scsiPrintMain(char *device, int fd);
+// MACROS to control printing behavior (from ataprint.h)
+#define QUIETON(control)  {if (control->quietmode) control->veryquietmode=0;}
+#define QUIETOFF(control) {if (control->quietmode && !control->veryquietmode) control->veryquietmode=1;}
+
+int scsiPrintMain(const char * dev_name, int fd);
 
 #endif
