@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-3 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,6 @@
  *
  */
 
-
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
@@ -37,22 +36,25 @@
 #include "atacmds.h"
 #include "ataprint.h"
 #include "extern.h"
+#include "knowndrives.h"
 #include "scsicmds.h"
 #include "scsiprint.h"
 #include "smartctl.h"
 #include "utility.h"
 
-extern const char *atacmds_c_cvsid, *ataprint_c_cvsid, *scsicmds_c_cvsid, *scsiprint_c_cvsid, *utility_c_cvsid; 
-const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.50 2003/01/21 03:17:33 ballen4705 Exp $"
-ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
+extern const char *atacmdnames_c_cvsid, *atacmds_c_cvsid, *ataprint_c_cvsid, *escalade_c_cvsid, *knowndrives_c_cvsid, *scsicmds_c_cvsid, *scsiprint_c_cvsid, *utility_c_cvsid; 
+const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.87 2003/08/18 07:10:41 ballen4705 Exp $"
+ATACMDS_H_CVSID ATAPRINT_H_CVSID EXTERN_H_CVSID KNOWNDRIVES_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // This is a block containing all the "control variables".  We declare
 // this globally in this file, and externally in other files.
-atamainctrl *con=NULL;
+smartmonctrl *con=NULL;
 
+// Track memory use
+long long bytes=0;
 
 void printslogan(){
-  pout("smartctl version %d.%d-%d Copyright (C) 2002 Bruce Allen\n",
+  pout("smartctl version %d.%d-%d Copyright (C) 2002-3 Bruce Allen\n",
       (int)RELEASE_MAJOR, (int)RELEASE_MINOR, (int)SMARTMONTOOLS_VERSION);
   pout("Home page is %s\n\n",PROJECTHOME);
   return;
@@ -66,10 +68,16 @@ void printcopy(){
   pout("under the terms of the GNU General Public License Version 2.\n");
   pout("See http://www.gnu.org for further details.\n\n");
   pout("CVS version IDs of files used to build this code are:\n");
+  printone(out,atacmdnames_c_cvsid);
+  pout("%s",out);
   printone(out,atacmds_c_cvsid);
   pout("%s",out);
   printone(out,ataprint_c_cvsid);
   pout("%s",out);
+  printone(out,escalade_c_cvsid);
+  pout("%s",out);
+  printone(out,knowndrives_c_cvsid);
+  pout("%s",out);
   printone(out,scsicmds_c_cvsid);
   pout("%s",out);
   printone(out,scsiprint_c_cvsid);
@@ -83,158 +91,184 @@ void printcopy(){
 
 /*  void prints help information for command syntax */
 void Usage (void){
-  printf("Usage: smartctl [options] [device]\n");
-  printf("\n==============  SHOW INFORMATION OPTIONS  =================================\n");
+  printf("Usage: smartctl [options] device\n\n");
+  printf("============================================ SHOW INFORMATION OPTIONS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
-  printf("\
-  -h, -?, --help, --usage\n\
-         Display this help and exit\n\
-  -V, --version, --copyright, --license\n\
-         Print license, copyright, and version information and exit\n\
-  -i, --info                                                       \n\
-         Show identity information for device\n\
-  -a, --all                                                        \n\
-         Show all SMART information for device\n\
-");
+  printf(
+"  -h, -?, --help, --usage\n"
+"         Display this help and exit\n\n"
+"  -V, --version, --copyright, --license\n"
+"         Print license, copyright, and version information and exit\n\n"
+"  -i, --info                                                       \n"
+"         Show identity information for device\n\n"
+"  -a, --all                                                        \n"
+"         Show all SMART information for device\n\n"
+  );
 #else
-  printf("\
-  -h, -?    Display this help and exit\n\
-  -V        Print license, copyright, and version information\n\
-  -i        Show identity information for device\n\
-  -a        Show all SMART information for device                   \n\
-");
+  printf(
+"  -h, -?    Display this help and exit\n"
+"  -V        Print license, copyright, and version information\n"
+"  -i        Show identity information for device\n"
+"  -a        Show all SMART information for device\n\n"
+  );
 #endif
-  printf("==============  SMARTCTL RUN-TIME BEHAVIOR OPTIONS  =======================\n");
+  printf("================================== SMARTCTL RUN-TIME BEHAVIOR OPTIONS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
-  printf("\
-  -q TYPE, --quietmode=TYPE                                           (ATA)\n\
-         Set smartctl quiet mode to one of: errorsonly, silent\n\
-  -d TYPE, --device=TYPE\n\
-         Specify device type to one of: ata, scsi\n\
-  -T TYPE, --tolerance=TYPE                                           (ATA)\n\
-         Set tolerance to one of: normal, conservative, permissive\n\
-  -b TYPE, --badsum=TYPE                                              (ATA)\n\
-         Set action on bad checksum to one of: warn, exit, ignore\n\
-");
+  printf(
+"  -q TYPE, --quietmode=TYPE                                           (ATA)\n"
+"         Set smartctl quiet mode to one of: errorsonly, silent\n\n"
+"  -d TYPE, --device=TYPE\n"
+"         Specify device type to one of: ata, scsi, 3ware,N\n\n"
+"  -T TYPE, --tolerance=TYPE                                           (ATA)\n"
+"         Set tolerance to one of: normal, conservative, permissive\n\n"
+"  -b TYPE, --badsum=TYPE                                              (ATA)\n"
+"         Set action on bad checksum to one of: warn, exit, ignore\n\n"
+"  -r TYPE, --report=TYPE\n"
+"         Report transactions (see man page)\n\n"
+  );
 #else
-  printf("\
-  -q TYPE   Set smartctl quiet mode to one of: errorsonly, silent     (ATA)\n\
-  -d TYPE   Specify device type to one of: ata, scsi\n\
-  -T TYPE   Set tolerance to one of: normal, conservative, permissive (ATA)\n\
-  -b TYPE   Set action on bad checksum to one of: warn, exit, ignore  (ATA)\n\
-");
+  printf(
+"  -q TYPE   Set smartctl quiet mode to one of: errorsonly, silent     (ATA)\n"
+"  -d TYPE   Specify device type to one of: ata, scsi, 3ware,N\n"
+"  -T TYPE   Set tolerance to one of: normal, conservative, permissive (ATA)\n"
+"  -b TYPE   Set action on bad checksum to one of: warn, exit, ignore  (ATA)\n"
+"  -r TYPE   Report transactions (see man page)\n\n"
+  );
 #endif
-  printf("==============  DEVICE FEATURE ENABLE/DISABLE COMMANDS  ===================\n");
+  printf("============================== DEVICE FEATURE ENABLE/DISABLE COMMANDS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
-  printf("\
-  -s VALUE, --smart=VALUE\n\
-        Enable/disable SMART on device (on/off)\n\
-  -o VALUE, --offlineauto=VALUE                                       (ATA)\n\
-        Enable/disable automatic offline testing on device (on/off)\n\
-  -S VALUE, --saveauto=VALUE                                          (ATA)\n\
-        Enable/disable Attribute autosave on device (on/off)\n\
-");
+  printf(
+"  -s VALUE, --smart=VALUE\n"
+"        Enable/disable SMART on device (on/off)\n\n"
+"  -o VALUE, --offlineauto=VALUE                                       (ATA)\n"
+"        Enable/disable automatic offline testing on device (on/off)\n\n"
+"  -S VALUE, --saveauto=VALUE                                          (ATA)\n"
+"        Enable/disable Attribute autosave on device (on/off)\n\n"
+  );
 #else
-  printf("\
-  -s VALUE  Enable/disable SMART on device (on/off)\n\
-  -o VALUE  Enable/disable device automatic offline testing (on/off)  (ATA)\n\
-  -S VALUE  Enable/disable device Attribute autosave (on/off)         (ATA)\n\
-");
+  printf(
+"  -s VALUE  Enable/disable SMART on device (on/off)\n"
+"  -o VALUE  Enable/disable device automatic offline testing (on/off)  (ATA)\n"
+"  -S VALUE  Enable/disable device Attribute autosave (on/off)         (ATA)\n\n"
+  );
 #endif
-  printf("==============  READ AND DISPLAY DATA OPTIONS  ============================\n");
+  printf("======================================= READ AND DISPLAY DATA OPTIONS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
-  printf("\
-  -H, --health\n\
-        Show device SMART health status\n\
-  -c, --capabilities                                                  (ATA)\n\
-        Show device SMART capabilities\n\
-  -A, --attributes                                                    (ATA)\n\
-        Show device SMART vendor-specific Attributes and values\n\
-  -l TYPE, --log=TYPE\n\
-        Show device log. Type is one of: error (ATA), selftest\n\
-  -v N,OPTION , --vendorattribute=N,OPTION                            (ATA)\n\
-        Set display OPTION for vendor Attribute N (see man page)\n\
-");
+  printf(
+"  -H, --health\n"
+"        Show device SMART health status\n\n"
+"  -c, --capabilities                                                  (ATA)\n"
+"        Show device SMART capabilities\n\n"
+"  -A, --attributes                                                    (ATA)\n"
+"        Show device SMART vendor-specific Attributes and values\n\n"
+"  -l TYPE, --log=TYPE\n"
+"        Show device log. Type is one of: error, selftest, directory\n\n"
+"  -v N,OPTION , --vendorattribute=N,OPTION                            (ATA)\n"
+"        Set display OPTION for vendor Attribute N (see man page)\n\n"
+"  -F TYPE, --firmwarebug=TYPE                                         (ATA)\n"
+"        Use firmware bug workaround. Type is one of: none, samsung\n\n"
+"  -P TYPE, --presets=TYPE                                             (ATA)\n"
+"        Drive-specific presets: use, ignore, show, showall\n\n"
+  );
 #else
-  printf("\
-  -H        Show device SMART health status\n\
-  -c        Show device SMART capabilities                            (ATA)\n\
-  -A        Show device SMART vendor-specific Attributes and values   (ATA)\n\
-  -l TYPE   Show device log. Type is one of: error (ATA), selftest\n\
-  -v N,OPT  Set display OPTion for vendor Attribute N (see man page)  (ATA)\n\
-");
+  printf(
+"  -H        Show device SMART health status\n"
+"  -c        Show device SMART capabilities                             (ATA)\n"
+"  -A        Show device SMART vendor-specific Attributes and values    (ATA)\n"
+"  -l TYPE   Show device log. Type is one of: error, selftest, directory\n"
+"  -v N,OPT  Set display OPTion for vendor Attribute N (see man page)   (ATA)\n"
+"  -F TYPE   Use firmware bug workaround. Type is one of: none, samsung (ATA)\n"
+"  -P TYPE   Drive-specific presets: use, ignore, show, showall         (ATA)\n\n"
+  );
 #endif
-  printf("==============  DEVICE SELF-TEST OPTIONS  =================================\n");
+  printf("============================================ DEVICE SELF-TEST OPTIONS =====\n\n");
 #ifdef HAVE_GETOPT_LONG
-  printf("\
-  -t TEST, --test=TEST\n\
-        Run test on device.  TEST is one of: offline, short, long\n\
-  -C, --captive\n\
-        With -t, performs test in captive mode (short/long only)\n\
-  -X, --abort\n\
-        Abort any non-captive test on device\n\
-");
+  printf(
+"  -t TEST, --test=TEST\n"
+#if DEVELOP_SELECTIVE_SELF_TEST
+"        Run test.  TEST is: offline, short, long, conveyance, selective\n\n"
 #else
-  printf("\
-  -t TEST   Run test on device.  TEST is one of: offline, short, long   \n\
-  -C        With -t, performs test in captive mode (short/long only)  \n\
-  -X        Abort any non-captive test                                \n\
-");
+"        Run test.  TEST is: offline, short, long, conveyance\n\n"
 #endif
-  printf("==============  SMARTCTL EXAMPLES  ========================================\n");
-#ifdef HAVE_GETOPT_LONG
-  printf("\
-  smartctl -a /dev/hda                 (Prints all SMART information)\n\
-  smartctl --smart=on --offlineauto=on --saveauto=on /dev/hda\n\
-                                       (Enables SMART on first disk)\n\
-  smartctl -t long /dev/hda            (Executes extended disk self-test)\n\
-  smartctl --attributes --log=selftest --quietmode=errorsonly /dev/hda\n\
-                                       (Prints Self-Test & Attribute errors)\n\
-");
+
+"  -C, --captive\n"
+#if DEVELOP_SELECTIVE_SELF_TEST
+"        With -t, do test in captive mode (short/long/conveyance/selective)\n\n"
 #else
-  printf("\
-  smartctl -a /dev/hda                 (Prints all SMART information)\n\
-  smartctl -s on -o on -S on /dev/hda  (Enables SMART on first disk)\n\
-  smartctl -t long /dev/hda            (Executes extended disk self-test)\n\
-  smartctl -A -l selftest -q errorsonly /dev/hda\n\
-                                       (Prints Self-Test & Attribute errors)\n\
-");
+"        With -t, do test in captive mode (short/long/conveyance)\n\n"
+#endif
+
+"  -X, --abort\n"
+"        Abort any non-captive test on device\n\n"
+);
+#else
+  printf(
+#if DEVELOP_SELECTIVE_SELF_TEST
+"  -t TEST   Run test.  TEST is: offline, short, long, conveyance, selective\n"
+"  -C        With -t, do test in captive mode (short/long/conveyance/selective)\n"
+#else
+"  -t TEST   Run test.  TEST is: offline, short, long, conveyance\n"
+"  -C        With -t, do test in captive mode (short/long/conveyance)\n"
+#endif
+"  -X        Abort any non-captive test\n\n"
+  );
+#endif
+  printf("=================================================== SMARTCTL EXAMPLES =====\n\n");
+#ifdef HAVE_GETOPT_LONG
+  printf(
+"  smartctl -a /dev/hda                       (Prints all SMART information)\n\n"
+"  smartctl --smart=on --offlineauto=on --saveauto=on /dev/hda\n"
+"                                              (Enables SMART on first disk)\n\n"
+"  smartctl -t long /dev/hda              (Executes extended disk self-test)\n\n"
+"  smartctl --attributes --log=selftest --quietmode=errorsonly /dev/hda\n"
+"                                      (Prints Self-Test & Attribute errors)\n"
+"  smartctl -a -device=3ware,2 /dev/sda\n"
+"          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
+  );
+#else
+  printf(
+"  smartctl -a /dev/hda                       (Prints all SMART information)\n"
+"  smartctl -s on -o on -S on /dev/hda         (Enables SMART on first disk)\n"
+"  smartctl -t long /dev/hda              (Executes extended disk self-test)\n"
+"  smartctl -A -l selftest -q errorsonly /dev/hda\n"
+"                                      (Prints Self-Test & Attribute errors)\n"
+"  smartctl -a -d 3ware,2 /dev/sda\n"
+"          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
+  );
 #endif
 }
 
 /* Returns a pointer to a static string containing a formatted list of the valid
-   arguments to the option opt or NULL on failure. */
+   arguments to the option opt or NULL on failure. Note 'v' case different */
 const char *getvalidarglist(char opt) {
-  static char *v_list = NULL;
-  char *s;
-
   switch (opt) {
   case 'q':
     return "errorsonly, silent";
   case 'd':
-    return "ata, scsi";
+    return "ata, scsi, 3ware,N";
   case 'T':
     return "normal, conservative, permissive";
   case 'b':
     return "warn, exit, ignore";
+  case 'r':
+    return "ioctl[,N], ataioctl[,N], scsiioctl[,N]";
   case 's':
   case 'o':
   case 'S':
     return "on, off";
   case 'l':
-    return "error, selftest";
-  case 'v':
-    if (v_list) 
-      return v_list;
-    if (!(s = create_vendor_attribute_arg_list()))
-      return NULL;
-    // Allocate space for "\"help\", " + s + terminating 0
-    v_list = (char *)malloc(9+strlen(s));
-    sprintf(v_list, "\"help\", %s", s);
-    free(s);
-    return v_list;
+    return "error, selftest, directory";
+  case 'P':
+    return "use, ignore, show, showall";
   case 't':
-    return "offline, short, long";
+#if DEVELOP_SELECTIVE_SELF_TEST
+    return "offline, short, long, conveyance, selective";
+#else
+    return "offline, short, long, conveyance";
+#endif
+  case 'F':
+    return "none, samsung";
+  case 'v':
   default:
     return NULL;
   }
@@ -243,16 +277,33 @@ const char *getvalidarglist(char opt) {
 /* Prints the message "=======> VALID ARGUMENTS ARE: <LIST>  <=======\n", where
    <LIST> is the list of valid arguments for option opt. */
 void printvalidarglistmessage(char opt) {
-  const char *s;
-
-  pout("=======> VALID ARGUMENTS ARE: ");
-  if (!(s = getvalidarglist(opt)))
-    pout("Error whilst constructing argument list for option %c", opt);
+  char *s;
+  
+  if (opt=='v')
+    s=create_vendor_attribute_arg_list();
   else
-    pout((char *)s);
-  pout(" <=======\n");
+    s=(char *)getvalidarglist(opt);
+  
+  if (!s) {
+    pout("Error whilst constructing argument list for option %c", opt);
+    return;
+  }
+ 
+  if (opt=='v'){
+    pout("=======> VALID ARGUMENTS ARE:\n\thelp\n%s\n<=======\n", s);
+    free(s);
+  }
+  else {
+  // getvalidarglist() might produce a multiline or single line string.  We
+  // need to figure out which to get the formatting right.
+    char separator = strchr(s, '\n') ? '\n' : ' ';
+    pout("=======> VALID ARGUMENTS ARE:%c%s%c<=======\n", separator, (char *)s, separator);
+  }
+
+  return;
 }
 
+
 unsigned char tryata=0,tryscsi=0;
 
 /*      Takes command options and sets features to be run */	
@@ -260,10 +311,11 @@ void ParseOpts (int argc, char** argv){
   int optchar;
   int badarg;
   int captive;
+  unsigned char *charp;
   extern char *optarg;
   extern int optopt, optind, opterr;
   // Please update getvalidarglist() if you edit shortopts
-  const char *shortopts = "h?Vq:d:T:b:s:o:S:HcAl:iav:t:CX";
+  const char *shortopts = "h?Vq:d:T:b:r:s:o:S:HcAl:iav:P:t:CXF:";
 #ifdef HAVE_GETOPT_LONG
   char *arg;
   // Please update getvalidarglist() if you edit longopts
@@ -277,6 +329,7 @@ void ParseOpts (int argc, char** argv){
     { "device",          required_argument, 0, 'd' },
     { "tolerance",       required_argument, 0, 'T' },
     { "badsum",          required_argument, 0, 'b' },
+    { "report",          required_argument, 0, 'r' },
     { "smart",           required_argument, 0, 's' },
     { "offlineauto",     required_argument, 0, 'o' },
     { "saveauto",        required_argument, 0, 'S' },
@@ -287,9 +340,11 @@ void ParseOpts (int argc, char** argv){
     { "info",            no_argument,       0, 'i' },
     { "all",             no_argument,       0, 'a' },
     { "vendorattribute", required_argument, 0, 'v' },
+    { "presets",         required_argument, 0, 'P' },
     { "test",            required_argument, 0, 't' },
     { "captive",         no_argument,       0, 'C' },
     { "abort",           no_argument,       0, 'X' },
+    { "firmwarebug",     required_argument, 0, 'F' },
     { 0,                 0,                 0, 0   }
   };
 #endif
@@ -327,14 +382,39 @@ void ParseOpts (int argc, char** argv){
       break;
     case 'd':
       if (!strcmp(optarg,"ata")) {
-        tryata  = TRUE;
-        tryscsi = FALSE;
+	tryata  = TRUE;
+	tryscsi = FALSE;
+	con->escalade = 0;
       } else if (!strcmp(optarg,"scsi")) {
-        tryata  = FALSE;
-        tryscsi = TRUE;
+	tryata  = FALSE;
+	tryscsi = TRUE;
+	con->escalade = 0;
       } else {
-        badarg = TRUE;
-      }
+	// look for RAID-type device
+	int i;
+        char *s;
+	
+	// make a copy of the string to mess with
+	if (!(s = strdup(optarg))) {
+          con->veryquietmode = FALSE;
+          pout("No memory for argument of -d. Exiting...\n");
+          exit(FAILCMD);
+        } else if (strncmp(s,"3ware,",6)) {
+	  badarg = TRUE;
+	} else if (split_report_arg2(s, &i)) {
+	  pout("Option -d 3ware,N requires N to be a non-negative integer\n");
+	  badarg = TRUE;
+        } else if (i<0 || i>15) {
+	  pout("Option -d 3ware,N (N=%d) must have 0 <= N <= 15\n", i);
+	  badarg = TRUE;
+	} else {
+	  // NOTE: escalade = disk number + 1
+          con->escalade = i+1;
+	  tryata  = TRUE;
+	  tryscsi = FALSE;
+	}
+        free(s);
+      } 	
       break;
     case 'T':
       if (!strcmp(optarg,"normal")) {
@@ -364,6 +444,33 @@ void ParseOpts (int argc, char** argv){
         badarg = TRUE;
       }
       break;
+    case 'r':
+      {
+        int i;
+        char *s;
+
+        // split_report_arg() may modify its first argument string, so use a
+        // copy of optarg in case we want optarg for an error message.
+        if (!(s = strdup(optarg))) {
+          con->veryquietmode = FALSE;
+          pout("Can't allocate memory to copy argument to -r option"
+               " - exiting\n");
+          exit(FAILCMD);
+        }
+        if (split_report_arg(s, &i)) {
+          badarg = TRUE;
+        } else if (!strcmp(s,"ioctl")) {
+          con->reportataioctl  = con->reportscsiioctl = i;
+        } else if (!strcmp(s,"ataioctl")) {
+          con->reportataioctl = i;
+        } else if (!strcmp(s,"scsiioctl")) {
+          con->reportscsiioctl = i;
+        } else {
+          badarg = TRUE;
+        }
+        free(s);
+      }
+      break;
     case 's':
       if (!strcmp(optarg,"on")) {
         con->smartenable  = TRUE;
@@ -400,6 +507,15 @@ void ParseOpts (int argc, char** argv){
     case 'H':
       con->checksmart = TRUE;		
       break;
+    case 'F':
+      if (!strcmp(optarg,"none")) {
+        con->fixfirmwarebug = FIX_NONE;
+      } else if (!strcmp(optarg,"samsung")) {
+        con->fixfirmwarebug = FIX_SAMSUNG;
+      } else {
+        badarg = TRUE;
+      }
+      break;
     case 'c':
       con->generalsmartvalues = TRUE;
       break;
@@ -411,6 +527,8 @@ void ParseOpts (int argc, char** argv){
         con->smarterrorlog = TRUE;
       } else if (!strcmp(optarg,"selftest")) {
         con->smartselftestlog = TRUE;
+      } else if (!strcmp(optarg,"directory")) {
+        con->smartlogdirectory = TRUE;
       } else {
         badarg = TRUE;
       }
@@ -436,13 +554,32 @@ void ParseOpts (int argc, char** argv){
           pout("Insufficient memory to construct argument list\n");
           exit(FAILCMD);
         }
-        pout("The valid arguments to -v are: \"help\", %s\n", s);
+        pout("The valid arguments to -v are:\n\thelp\n%s\n", s);
         free(s);
         exit(0);
       }
-      if (parse_attribute_def(optarg, con->attributedefs))
+      charp=con->attributedefs;
+      if (!charp){
+	pout("Fatal internal error in ParseOpts()\n");
+	exit(FAILCMD);
+      }
+      if (parse_attribute_def(optarg, &charp))
 	badarg = TRUE;
       break;    
+    case 'P':
+      if (!strcmp(optarg, "use")) {
+        con->ignorepresets = FALSE;
+      } else if (!strcmp(optarg, "ignore")) {
+        con->ignorepresets = TRUE;
+      } else if (!strcmp(optarg, "show")) {
+        con->showpresets = TRUE;
+      } else if (!strcmp(optarg, "showall")) {
+        showallpresets();
+        exit(0);
+      } else {
+        badarg = TRUE;
+      }
+      break;
     case 't':
       if (!strcmp(optarg,"offline")) {
         con->smartexeoffimmediate = TRUE;
@@ -453,6 +590,14 @@ void ParseOpts (int argc, char** argv){
       } else if (!strcmp(optarg,"long")) {
         con->smartextendselftest = TRUE;
         con->testcase            = EXTEND_SELF_TEST;
+      } else if (!strcmp(optarg,"conveyance")) {
+        con->smartconveyanceselftest = TRUE;
+        con->testcase            = CONVEYANCE_SELF_TEST;
+#if DEVELOP_SELECTIVE_SELF_TEST
+      } else if (!strcmp(optarg,"selective")) {
+        con->smartselectiveselftest = TRUE;
+        con->testcase            = SELECTIVE_SELF_TEST;
+#endif
       } else {
         badarg = TRUE;
       }
@@ -515,7 +660,7 @@ void ParseOpts (int argc, char** argv){
   // Do this here, so results are independent of argument order	
   if (con->quietmode)
     con->veryquietmode=TRUE;
-  
+
   // error message if user has asked for more than one test
   if (1<(con->smartexeoffimmediate+con->smartshortselftest+con->smartextendselftest+
 	 con->smartshortcapselftest+con->smartextendcapselftest+con->smartselftestabort)){
@@ -528,15 +673,27 @@ void ParseOpts (int argc, char** argv){
 
   // If captive option was used, change test type if appropriate.
   if (captive && con->smartshortselftest) {
-      con->smartshortselftest    = FALSE;
-      con->smartshortcapselftest = TRUE;
-      con->testcase              = SHORT_CAPTIVE_SELF_TEST;
+    con->smartshortselftest    = FALSE;
+    con->smartshortcapselftest = TRUE;
+    con->testcase              = SHORT_CAPTIVE_SELF_TEST;
   } else if (captive && con->smartextendselftest) {
-      con->smartextendselftest    = FALSE;
-      con->smartextendcapselftest = TRUE;
-      con->testcase               = EXTEND_CAPTIVE_SELF_TEST;
+    con->smartextendselftest    = FALSE;
+    con->smartextendcapselftest = TRUE;
+    con->testcase               = EXTEND_CAPTIVE_SELF_TEST;
   }
-
+  else if (captive && con->smartconveyanceselftest) {
+    con->smartconveyanceselftest    = FALSE;
+    con->smartconveyancecapselftest = TRUE;
+    con->testcase                   = CONVEYANCE_CAPTIVE_SELF_TEST;
+  }
+#if DEVELOP_SELECTIVE_SELF_TEST
+  else if (captive && con->smartselectiveselftest) {
+    con->smartselectiveselftest    = FALSE;
+    con->smartselectivecapselftest = TRUE;
+    con->testcase                  = SELECTIVE_CAPTIVE_SELF_TEST;
+  }
+#endif 
+ 
   // From here on, normal operations...
   printslogan();
   
@@ -579,6 +736,7 @@ void pout(char *fmt, ...){
   // print out
   vprintf(fmt,ap);
   va_end(ap);
+  fflush(stdout);
   return;
 }
 
@@ -587,8 +745,8 @@ void pout(char *fmt, ...){
 int main (int argc, char **argv){
   int fd,retval=0;
   char *device;
-  atamainctrl control;
-  const char *devroot="/dev/";
+  smartmonctrl control;
+  int dev_type, flags;
 
   // define control block for external functions
   con=&control;
@@ -596,8 +754,21 @@ int main (int argc, char **argv){
   // Part input arguments
   ParseOpts(argc,argv);
 
-  // open device - read-only mode is enough to issue needed commands
-  fd = open(device=argv[argc-1], O_RDONLY);  
+  device = argv[argc-1];
+  dev_type = guess_linux_device_type(device);
+  if (GUESS_DEVTYPE_SCSI == dev_type) {
+    flags = O_RDWR | O_NONBLOCK;
+    tryscsi = 1;
+  } else { /* treat "don't know" case as an ATA device as well */
+    flags = O_RDONLY | O_NONBLOCK;
+    tryata = 1;
+  }
+
+  // open device - SCSI devices are opened (O_RDWR | O_NONBLOCK) so the
+  // scsi generci device can be used (needs write permission for MODE 
+  // SELECT command) plus O_NONBLOCK to stop open hanging if media not
+  // present (e.g. with st). 
+  fd = open(device, flags);
   if (fd<0) {
     char errmsg[256];
     snprintf(errmsg,256,"Smartctl open device: %s failed",argv[argc-1]);
@@ -606,27 +777,11 @@ int main (int argc, char **argv){
     return FAILDEV;
   }
 
-  // if necessary, try to guess if this is an ATA or SCSI device
-  if (!tryata && !tryscsi) {
-    if (!strncmp(device,devroot,strlen(devroot)) && strlen(device)>5){
-      if (device[5] == 'h')
-	tryata=1;
-      if (device[5] == 's')
-	tryscsi=1;
-    }
-    else if (strlen(device)){
-	if (device[0] == 'h')
-	  tryata=1;
-	if (device[0] == 's')
-	  tryscsi=1;
-      }
-  }
-
   // now call appropriate ATA or SCSI routine
   if (tryata)
-    retval=ataPrintMain(fd);
+    retval = ataPrintMain(fd);
   else if (tryscsi)
-    scsiPrintMain (device, fd);
+    retval = scsiPrintMain(device, fd);
   else {
     pout("Smartctl: specify if this is an ATA or SCSI device with the -d option.\n");
     Usage();
